@@ -34,7 +34,7 @@ THE SOFTWARE.
 typedef struct {
 	PianoHandle_t ph;
 	WaitressHandle_t waith;
-	struct audioPlayer player;
+	player_t player;
 	BarSettings_t settings;
 	/* first item is current song */
 	PianoSong_t *playlist;
